import { gettext, InjectionManager, Extension } from 'resource:///org/gnome/shell/extensions/extension.js';
import GLib from 'gi://GLib';
import GObject from 'gi://GObject';
import Gio from 'gi://Gio';
import Meta from 'gi://Meta';
import * as Main from 'resource:///org/gnome/shell/ui/main.js';
import Clutter from 'gi://Clutter';
import Pango from 'gi://Pango';
import St from 'gi://St';
import * as Params from 'resource:///org/gnome/shell/misc/params.js';
import * as PanelMenu from 'resource:///org/gnome/shell/ui/panelMenu.js';
import * as PopupMenu from 'resource:///org/gnome/shell/ui/popupMenu.js';
import * as BoxPointer from 'resource:///org/gnome/shell/ui/boxpointer.js';
import * as KeyboardBase from 'resource:///org/gnome/shell/ui/keyboard.js';
import * as InputSourceManager from 'resource:///org/gnome/shell/ui/status/keyboard.js';
import Mtk from 'gi://Mtk';

const KimMenuItem = GObject.registerClass(class KimMenuItem extends PopupMenu.PopupBaseMenuItem {
    constructor(text, iconName, params) {
        super(params);
        this.label = new St.Label({ text });
        this._icon = new St.Icon({
            style_class: "popup-menu-icon",
            x_align: Clutter.ActorAlign.END,
        });
        this.add_child(this._icon);
        this.add_child(this.label);
        this.setIcon(iconName);
    }
    setIcon(name) {
        const icon = createIcon(name);
        if (icon != null)
            this._icon.gicon = icon;
    }
});
const parseProperty = (str) => {
    const p = str.split(":");
    const property = {
        hint: p.length > 4 && p[4].length > 0 ? p[4].split(",") : [],
        icon: p[2],
        key: p[0],
        label: p[1],
        text: p[3],
    };
    return property;
};
const createIcon = (name) => {
    if (!name)
        return undefined;
    if (name[0] == "/") {
        return Gio.FileIcon.new(Gio.File.new_for_path(name));
    }
    return Gio.ThemedIcon.new_with_default_fallbacks(`${name}-symbolic-hack`);
};
const createMenuItem = (property) => {
    const item = new KimMenuItem("", "");
    item.key = property.key;
    return item;
};
const getTextStyle = (settings) => {
    const fontString = settings?.get_string("font") || "Sans 11";
    const desc = Pango.FontDescription.from_string(fontString);
    const fontFamily = desc.get_family();
    const fontSize = `${desc.get_size() / Pango.SCALE}pt`;
    let fontStyle = "normal";
    for (const style in Pango.Style)
        if (Pango.Style[style] === desc.get_style()) {
            fontStyle = style.toLowerCase();
            break;
        }
    let fontWeight = Pango.Weight.NORMAL;
    try {
        fontWeight = desc.get_weight();
    }
    catch (_error) {
    }
    return `font-family:'${fontFamily}';font-size:${fontSize};font-style:${fontStyle};font-weight:${fontWeight}`;
};
const keyboardIsVisible = () => {
    return Main.keyboard._keyboard?.visible ?? false;
};
const isLookupTableVertical = (settings) => {
    return settings?.get_boolean("vertical") ?? false;
};
const extractLabelString = (l) => {
    if (l.length >= 2 && l.charCodeAt(0) < 127 && l.charCodeAt(1) < 127) {
        return l.substring(0, 2);
    }
    return l.substring(0, 1);
};

const KimIndicator = GObject.registerClass(class KimIndicator extends PanelMenu.Button {
    constructor(params) {
        super(0.5, "kimpanel");
        const _params = Params.parse(params, { kimpanel: null });
        this.properties = {};
        this.propertySwitch = {};
        const hbox = new St.BoxLayout({ style_class: "panel-status-menu-box" });
        this.labelIcon = new St.Label({ y_align: Clutter.ActorAlign.CENTER });
        this.labelIcon.clutter_text.ellipsize = Pango.EllipsizeMode.NONE;
        this.labelIcon.clutter_text.line_wrap = false;
        this.mainIcon = new St.Icon({
            gicon: createIcon("input-keyboard"),
            style_class: "system-status-icon",
        });
        hbox.add_child(this.labelIcon);
        hbox.add_child(this.mainIcon);
        this.add_child(hbox);
        this.deactive();
        this.kimpanel = _params.kimpanel;
        const settingMenu = new PopupMenu.PopupMenuItem(gettext("Settings"));
        settingMenu.connect("activate", () => this.kimpanel?.emit("Configure"));
        const reloadMenu = new PopupMenu.PopupMenuItem(gettext("Reload Configuration"));
        reloadMenu.connect("activate", () => this.kimpanel?.emit("ReloadConfig"));
        if ("addMenuItem" in this.menu) {
            this.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
            this.menu.addMenuItem(reloadMenu);
            this.menu.addMenuItem(settingMenu);
        }
        this.hide();
    }
    active() {
        if (this.properties["/Fcitx/im"]) {
            this.setIcon(this.properties["/Fcitx/im"]);
        }
        else {
            this.deactive();
        }
    }
    deactive() {
        const property = {
            hint: [],
            icon: "input-keyboard",
            key: "",
            label: "",
            text: "",
        };
        this.setIcon(property);
    }
    destroy() {
        this.kimpanel = null;
        this.labelIcon.destroy();
        this.mainIcon.destroy();
        super.destroy();
    }
    updateProperties(properties) {
        if (properties == null) {
            for (const key in this.propertySwitch) {
                const property = this.properties[key];
                const item = this.propertySwitch[key];
                item.setIcon(property.icon);
                item.label.text = property.label;
            }
            return;
        }
        const _properties = Object.fromEntries(properties.map((value) => {
            const [key, ...frags] = value.split(":");
            return [key, frags.join(":")];
        }));
        for (const p in this.propertySwitch) {
            if (!(p in _properties) || _properties[p] == null) {
                this.propertySwitch[p].destroy();
                delete this.propertySwitch[p];
            }
        }
        let count = 0;
        for (const _property of properties) {
            const property = parseProperty(_property);
            if (property == null) {
                continue;
            }
            count++;
            const key = property.key;
            this.properties[key] = property;
            if (key in this.propertySwitch)
                this.updatePropertyItem(key);
            else
                this.addPropertyItem(key);
        }
        if (count > 0) {
            this.show();
        }
        else {
            this.hide();
        }
    }
    updateProperty(propstr) {
        const property = parseProperty(propstr);
        if (property == null) {
            return;
        }
        const key = property.key;
        this.properties[key] = property;
        this.updateProperties();
    }
    addPropertyItem(key) {
        if (!(key in this.properties)) {
            return;
        }
        const property = this.properties[key];
        const item = createMenuItem(property);
        item.menuItemActivateId = item.connect("activate", () => this.kimpanel?.triggerProperty(item.key));
        item.menuItemDestroyId = item.connect("destroy", () => {
            item.disconnect(item.menuItemActivateId);
            item.disconnect(item.menuItemDestroyId);
        });
        item.setIcon(property.icon);
        item.label.text = property.label;
        this.propertySwitch[key] = item;
        if ("addMenuItem" in this.menu)
            this.menu.addMenuItem(this.propertySwitch[key], this.menu.numMenuItems - 3);
    }
    setIcon(property) {
        for (const hint of property.hint) {
            if (hint.startsWith("label=")) {
                const label = hint.slice(6);
                if (label.length > 0) {
                    this.labelIcon.text = extractLabelString(label);
                    this.mainIcon.visible = false;
                    this.labelIcon.visible = true;
                    return;
                }
            }
        }
        const iconName = property.icon;
        const labelName = property.label;
        if (iconName === "") {
            this.labelIcon.text = extractLabelString(labelName);
            this.mainIcon.visible = false;
            this.labelIcon.visible = true;
        }
        else {
            const gicon = createIcon(iconName);
            if (gicon != null)
                this.mainIcon.gicon = gicon;
            this.mainIcon.visible = true;
            this.labelIcon.visible = false;
        }
    }
    updatePropertyItem(key) {
        const property = this.properties[key];
        const item = this.propertySwitch[key];
        item.setIcon(property.icon);
        item.label.text = property.label;
    }
});

const KEY_LONG_PRESS_TIME = 250;
const AllSuggestions = GObject.registerClass(class AllSuggestions extends St.ScrollView {
    kimpanel;
    constructor(kimpanel) {
        super({
            hscrollbarPolicy: St.PolicyType.NEVER,
            overlay_scrollbars: true,
            reactive: true,
            vscrollbarPolicy: St.PolicyType.AUTOMATIC,
            xExpand: true,
            yExpand: true,
        });
        this.kimpanel = kimpanel;
        this.boxLayout = new St.BoxLayout({
            styleClass: "word-suggestions word-all-suggestions",
            vertical: true,
            xExpand: true,
            yExpand: true,
        });
        this.gesture = new Clutter.GestureAction();
        this.gesture.connect("gesture-progress", (action) => {
            const [, , dy] = action.get_motion_delta(0);
            const adjustment = this.get_vadjustment();
            adjustment.value -= dy;
            return Clutter.EVENT_STOP;
        });
        this.pressTimeoutId = 0;
        this.add_child(this.boxLayout);
        this.add_action(this.gesture);
    }
    destroy() {
        if (this.boxLayout != null)
            this.remove_child(this.boxLayout);
        this.boxLayout = null;
        if (this.gesture != null)
            this.remove_action(this.gesture);
        this.gesture = null;
        if (this.pressTimeoutId !== 0)
            GLib.Source.remove(this.pressTimeoutId);
        this.pressTimeoutId = 0;
        super.destroy();
    }
    reset() {
        this.boxLayout?.remove_all_children();
        this.hide();
    }
    set(texts) {
        this.boxLayout?.remove_all_children();
        this.show();
        for (const text of texts) {
            const row = this.getRow();
            const button = new St.Button({
                label: text,
            });
            const callback = () => {
                this.kimpanel.selectCandidateText(text);
                Main.keyboard._keyboard?._aspectContainer?.show();
                this.hide();
            };
            button.connect("button-press-event", () => {
                this.buttonPress();
                return Clutter.EVENT_PROPAGATE;
            });
            button.connect("button-release-event", () => {
                this.buttonRelease(callback);
                return Clutter.EVENT_STOP;
            });
            button.connect("touch-event", (_actor, event) => {
                if (event.type() === Clutter.EventType.TOUCH_BEGIN) {
                    this.buttonPress();
                }
                else if (event.type() === Clutter.EventType.TOUCH_END) {
                    this.buttonRelease(callback);
                }
                return Clutter.EVENT_STOP;
            });
            row.add_child(button);
            if (row.width > (Main.keyboard._keyboard?.width ?? 0)) {
                row.remove_child(button);
                const newRow = new St.BoxLayout({ vertical: false });
                newRow.add_child(button);
                this.boxLayout?.add_child(newRow);
            }
        }
    }
    buttonPress() {
        this.pressTimeoutId = GLib.timeout_add(GLib.PRIORITY_DEFAULT, KEY_LONG_PRESS_TIME, () => {
            this.pressTimeoutId = 0;
            return GLib.SOURCE_REMOVE;
        });
    }
    buttonRelease(callback) {
        if (this.pressTimeoutId !== 0) {
            callback();
            GLib.source_remove(this.pressTimeoutId);
            this.pressTimeoutId = 0;
        }
    }
    getRow() {
        const row = this.boxLayout?.get_last_child();
        if (row != null)
            return row;
        const newRow = new St.BoxLayout({ vertical: false });
        this.boxLayout?.add_child(newRow);
        return newRow;
    }
});
const CustomKey = GObject.registerClass({
    Signals: {
        commit: { param_types: [GObject.TYPE_STRING] },
        keyval: { param_types: [GObject.TYPE_UINT] },
        "long-press": {},
        pressed: {},
        released: {},
    },
}, class CustomKey extends St.BoxLayout {
    constructor(params, extendedKeys = []) {
        const { label, iconName, commitString, keyval } = {
            keyval: 0,
            ...params,
        };
        super({ style_class: "key-container" });
        this.keyval =
            typeof keyval === "number" ? keyval : Number.parseInt(keyval, 16);
        this.keyButton = this.makeKey(commitString, label, iconName);
        this.add_child(this.keyButton);
        this.connect("destroy", this.onDestroy.bind(this));
        this.extendedKeys = extendedKeys;
        this.extendedKeyboard = null;
        this.pressTimeoutId = 0;
        this.capturedPress = false;
    }
    get iconName() {
        return this.icon.icon_name ?? "";
    }
    get subkeys() {
        return this.boxPointer;
    }
    set iconName(value) {
        this.icon.icon_name = value;
    }
    setLatched(latched) {
        if (latched)
            this.keyButton?.add_style_pseudo_class("latched");
        else
            this.keyButton?.remove_style_pseudo_class("latched");
    }
    cancel() {
        if (this.pressTimeoutId !== 0) {
            GLib.source_remove(this.pressTimeoutId);
            this.pressTimeoutId = 0;
        }
        this.touchPressSlot = null;
        this.keyButton?.set_hover(false);
        this.keyButton?.fake_release();
    }
    ensureExtendedKeysPopup() {
        if (this.extendedKeys.length === 0)
            return;
        if (this?.boxPointer)
            return;
        const _boxPointer = new BoxPointer.BoxPointer(St.Side.BOTTOM);
        this.boxPointer = _boxPointer;
        _boxPointer.hide();
        Main.layoutManager.addTopChrome(_boxPointer);
        if (this.keyButton)
            _boxPointer.setPosition(this.keyButton, 0.5);
        _boxPointer.add_style_class_name("keyboard-subkeys");
        this.getExtendedKeys();
        if (this.keyButton)
            this.keyButton._extendedKeys = this.extendedKeyboard;
    }
    getExtendedKeys() {
        this.extendedKeyboard = new St.BoxLayout({
            style_class: "key-container",
            vertical: false,
        });
        for (const extendedKey of this.extendedKeys) {
            const key = this.makeKey(extendedKey);
            key.extendedKey = extendedKey;
            this.extendedKeyboard?.add_child(key);
            if (this.keyButton) {
                const keyButton = this.keyButton;
                key.set_size(...keyButton.allocation.get_size());
                keyButton.connect("notify::allocation", () => key.set_size(...keyButton.allocation.get_size()));
            }
        }
        const _extendedKeyboard = this.extendedKeyboard;
        if (_extendedKeyboard != null)
            this.boxPointer?.bin.add_child(_extendedKeyboard);
    }
    hideSubkeys() {
        if (this.boxPointer)
            this.boxPointer?.close(BoxPointer.PopupAnimation.FULL);
        this.keyButton?.remove_style_class_name("active");
        global.stage.disconnectObject(this);
        this.keyButton?.disconnectObject(this);
        this.capturedPress = false;
    }
    makeKey(commitString, label, icon) {
        const button = new St.Button({
            style_class: "keyboard-key",
            x_expand: true,
        });
        if (icon) {
            const child = new St.Icon({ icon_name: icon });
            button.set_child(child);
            this.icon = child;
        }
        else if (label) {
            button.set_label(label);
        }
        else if (commitString) {
            button.set_label(commitString);
        }
        button.connect("button-press-event", () => {
            this.press(button);
            button.add_style_pseudo_class("active");
            return Clutter.EVENT_STOP;
        });
        button.connect("button-release-event", () => {
            this.release(button, commitString);
            button.remove_style_pseudo_class("active");
            return Clutter.EVENT_STOP;
        });
        button.connect("touch-event", (_actor, event) => {
            if (!Meta.is_wayland_compositor())
                return Clutter.EVENT_PROPAGATE;
            const slot = event.get_event_sequence().get_slot();
            if (!this.touchPressSlot &&
                event.type() === Clutter.EventType.TOUCH_BEGIN) {
                this.touchPressSlot = slot;
                this.press(button);
                button.add_style_pseudo_class("active");
            }
            else if (event.type() === Clutter.EventType.TOUCH_END) {
                if (!this.touchPressSlot || this.touchPressSlot === slot) {
                    this.release(button, commitString);
                    button.remove_style_pseudo_class("active");
                }
                if (this.touchPressSlot === slot)
                    this.touchPressSlot = null;
            }
            return Clutter.EVENT_STOP;
        });
        return button;
    }
    onCapturedEvent(_actor, event) {
        const type = event.type();
        const press = type === Clutter.EventType.BUTTON_PRESS ||
            type === Clutter.EventType.TOUCH_BEGIN;
        const release = type === Clutter.EventType.BUTTON_RELEASE ||
            type === Clutter.EventType.TOUCH_END;
        const targetActor = global.stage.get_event_actor(event);
        if (targetActor &&
            (targetActor === this.boxPointer?.bin ||
                this.boxPointer?.bin.contains(targetActor)))
            return Clutter.EVENT_PROPAGATE;
        if (press)
            this.capturedPress = true;
        else if (release && this.capturedPress) {
            this.hideSubkeys();
        }
        return Clutter.EVENT_STOP;
    }
    onDestroy() {
        if (this.boxPointer) {
            this.boxPointer?.destroy();
            this.boxPointer = null;
        }
        this.cancel();
    }
    press(button) {
        if (button === this.keyButton) {
            this.pressTimeoutId = GLib.timeout_add(GLib.PRIORITY_DEFAULT, KEY_LONG_PRESS_TIME, () => {
                this.pressTimeoutId = 0;
                this.emit("long-press");
                if (this.extendedKeys.length > 0) {
                    this.touchPressSlot = null;
                    this.ensureExtendedKeysPopup();
                    this.keyButton?.set_hover(false);
                    this.keyButton?.fake_release();
                    this.showSubkeys();
                }
                return GLib.SOURCE_REMOVE;
            });
        }
        this.emit("pressed");
        this.pressed = true;
    }
    release(button, commitString) {
        if (this.pressTimeoutId !== 0) {
            GLib.source_remove(this.pressTimeoutId);
            this.pressTimeoutId = 0;
        }
        if (this.pressed) {
            if (this.keyval && button === this.keyButton)
                this.emit("keyval", this.keyval);
            else if (commitString)
                this.emit("commit", commitString);
            else
                console.error("Need keyval or commitString");
        }
        this.emit("released");
        this.hideSubkeys();
        this.pressed = false;
    }
    showSubkeys() {
        this.boxPointer?.open(BoxPointer.PopupAnimation.FULL);
        global.stage.connectObject("captured-event", this.onCapturedEvent.bind(this), this);
        this.keyButton?.connectObject("notify::mapped", () => {
            if (!this.keyButton || !this.keyButton.is_mapped())
                this.hideSubkeys();
        }, this);
    }
});
const ExpandButton = GObject.registerClass(class ExpandButton extends St.Button {
    constructor() {
        super({
            iconName: "pan-down-symbolic",
            styleClass: "kimpanel-suggestions-expand",
        });
    }
    expand(expanded) {
        this.iconName = expanded ? "pan-up-symbolic" : "pan-down-symbolic";
    }
});
class LanguageSelectionPopup extends PopupMenu.PopupMenu {
    constructor(sourceActor) {
        super(sourceActor, 0.5, St.Side.BOTTOM);
        const inputSourceManager = InputSourceManager.getInputSourceManager();
        const inputSources = inputSourceManager.inputSources;
        let item;
        for (const i in inputSources) {
            const is = inputSources[i];
            item = this.addAction(is.displayName, () => {
                inputSourceManager.activateInputSource(is, true);
            });
            item.can_focus = false;
            item.setOrnament(is === inputSourceManager.currentSource
                ? PopupMenu.Ornament.DOT
                : PopupMenu.Ornament.NO_DOT);
        }
        this.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());
        item = this.addSettingsAction(gettext("Keyboard Settings"), "gnome-keyboard-panel.desktop");
        item.can_focus = false;
        sourceActor.connectObject("notify::mapped", () => {
            if (!sourceActor.is_mapped())
                this.close(true);
        }, this);
    }
    _onCapturedEvent(_actor, event) {
        const type = event.type();
        const press = type === Clutter.EventType.BUTTON_PRESS ||
            type === Clutter.EventType.TOUCH_BEGIN;
        const release = type === Clutter.EventType.BUTTON_RELEASE ||
            type === Clutter.EventType.TOUCH_END;
        const targetActor = global.stage.get_event_actor(event);
        if (targetActor &&
            (targetActor === this.actor || this.actor.contains(targetActor)))
            return Clutter.EVENT_PROPAGATE;
        if (press)
            this.sourceActor.capturedPress = true;
        else if (release && this.sourceActor.capturedPress) {
            this.close(true);
        }
        return Clutter.EVENT_STOP;
    }
    close(animate) {
        super.close(animate);
        this.sourceActor.capturedPress = false;
        global.stage.disconnectObject(this);
    }
    destroy() {
        global.stage.disconnectObject(this);
        this.sourceActor.disconnectObject(this);
        super.destroy();
    }
    open(animate) {
        super.open(animate);
        global.stage.connectObject("captured-event", this._onCapturedEvent.bind(this), this);
    }
}
const Keyboard = GObject.registerClass(class Keyboard extends GObject.Object {
    kimpanel;
    _dir;
    constructor(kimpanel, _dir) {
        super();
        this.kimpanel = kimpanel;
        this._dir = _dir;
        this.injectionManager = new InjectionManager();
        this.kanaActive = false;
        this.toggleIMKeySet = new Set();
        this.setupKeyboard();
    }
    destroy() {
        this.injectionManager?.clear();
        this.injectionManager = null;
        this.toggleIMKeySet = null;
        Main.layoutManager.removeChrome(Main.layoutManager.keyboardBox);
        const destroyed = this.destroyKeyboard();
        this.getModifiedLayouts()?._unregister();
        this.getDefaultLayouts()._register();
        if (destroyed) {
            this.allSuggestions?.destroy();
            this.allSuggestions = null;
            Main.keyboard._keyboard = new KeyboardBase.Keyboard();
        }
        Main.layoutManager.addTopChrome(Main.layoutManager.keyboardBox);
    }
    setSuggestions(texts) {
        Main.keyboard.resetSuggestions();
        Main.keyboard._keyboard?._suggestions?.set_width(-1);
        Main.keyboard._keyboard?._aspectContainer?.show();
        this.allSuggestions?.reset();
        const containerWidth = Main.keyboard._keyboard?.width ?? 0;
        for (const text of texts) {
            Main.keyboard.addSuggestion(text, () => {
                this.kimpanel.selectCandidateText(text);
            });
            const width = Main.keyboard._keyboard?._suggestions?.width ?? 0;
            if (width > containerWidth) {
                Main.keyboard._keyboard?._suggestions?.remove_child(Main.keyboard._keyboard._suggestions.lastChild);
                break;
            }
        }
        const suggestionsCount = Main.keyboard._keyboard?._suggestions?.get_children().length ?? 0;
        if (suggestionsCount === texts.length)
            return;
        const button = new ExpandButton();
        const spacer = new St.Widget({ x_expand: true });
        Main.keyboard._keyboard?._suggestions?.add_child(spacer);
        Main.keyboard._keyboard?._suggestions?.add_child(button);
        if (Main.keyboard._keyboard?._suggestions == null)
            return;
        while (Main.keyboard._keyboard._suggestions.width > containerWidth) {
            const key = Main.keyboard._keyboard._suggestions.get_child_at_index(Main.keyboard._keyboard._suggestions.get_children().length - 3);
            if (key != null) {
                Main.keyboard._keyboard?._suggestions.remove_child(key);
            }
        }
        const _suggestions = texts.slice(Main.keyboard._keyboard._suggestions.get_children().length - 2);
        const callback = () => {
            if (Main.keyboard._keyboard?._aspectContainer?.visible) {
                Main.keyboard._keyboard?._aspectContainer?.hide();
                this.allSuggestions?.set(_suggestions);
                button.expand(true);
            }
            else {
                Main.keyboard._keyboard?._aspectContainer?.show();
                this.allSuggestions?.reset();
                button.expand(false);
            }
        };
        button.connect("button-press-event", () => {
            callback();
            return Clutter.EVENT_STOP;
        });
        button.connect("touch-event", (_actor, event) => {
            if (event.type() !== Clutter.EventType.TOUCH_BEGIN)
                return Clutter.EVENT_PROPAGATE;
            callback();
            return Clutter.EVENT_STOP;
        });
        Main.keyboard._keyboard._suggestions.set_width(containerWidth);
        Main.keyboard._keyboard._suggestions.set_x_align(Clutter.ActorAlign.START);
    }
    updateProperty(value) {
        if (this.toggleIMKeySet == null)
            return;
        const kanaActive = /^\/Fcitx\/im:(?:Anthy:fcitx-anthy:ひらがな|Mozc:fcitx-mozc:全角かな|Mozc:fcitx_mozc_hiragana:全角かな|SKK:fcitx_skk:ひらがな|かな漢字:fcitx_kkc:ひらがな)/.test(value);
        this.kanaActive = kanaActive;
        for (const button of this.toggleIMKeySet.values()) {
            button.setLatched(kanaActive);
        }
        return;
    }
    destroyKeyboard() {
        try {
            Main.keyboard.keyboardActor?.destroy();
            Main.keyboard._keyboard = null;
        }
        catch (e) {
            if (e instanceof TypeError)
                return false;
            throw e;
        }
        return true;
    }
    getDefaultLayouts() {
        return Gio.Resource.load("/usr/share/gnome-shell/gnome-shell-osk-layouts.gresource");
    }
    getModifiedLayouts() {
        const modifiedLayoutsPath = this._dir
            ?.get_child("data")
            .get_child("gnome-shell-osk-layouts.gresource")
            .get_path();
        return modifiedLayoutsPath == null
            ? null
            : Gio.Resource.load(modifiedLayoutsPath);
    }
    overrideAddRowKeys(_originalMethod) {
        const _this = this;
        return function (keys, layout, emojiVisible) {
            let accumulatedWidth = 0;
            for (const key of keys) {
                const { strings } = key;
                const commitString = strings?.shift();
                if (key.action === "emoji" && !emojiVisible) {
                    accumulatedWidth = key.width ?? 1;
                    continue;
                }
                if (accumulatedWidth > 0) {
                    key.width = (key.width ?? 1) + accumulatedWidth;
                    accumulatedWidth = 0;
                }
                const button = new CustomKey({
                    commitString,
                    iconName: key.iconName,
                    keyval: key.keyval,
                    label: key.label,
                }, strings);
                if (key.keyval) {
                    button.connect("keyval", (_actor, keyval) => {
                        this._keyboardController.keyvalPress(keyval);
                        this._keyboardController.keyvalRelease(keyval);
                    });
                }
                if (key.action !== "modifier") {
                    button.connect("commit", (_actor, str) => {
                        this._keyboardController
                            .commit(str, this._modifiers)
                            .then(() => {
                            this._disableAllModifiers();
                            if (layout.mode === "default" ||
                                (layout.mode === "latched" && !this._latched)) {
                                if (this._contentHint !== 0)
                                    this._updateLevelFromHints();
                                else
                                    this._setActiveLevel("default");
                            }
                        })
                            .catch(console.error);
                    });
                }
                if (key.action != null) {
                    button.connect("released", () => {
                        if (key.action === "hide") {
                            this.close(true);
                        }
                        else if (key.action === "languageMenu") {
                            _this.kimpanel.toggleIM();
                        }
                        else if (key.action === "emoji") {
                            this._toggleEmoji();
                        }
                        else if (key.action === "modifier") {
                            if (key.keyval)
                                this._toggleModifier(key.keyval);
                        }
                        else if (key.action === "delete") {
                            this._keyboardController.toggleDelete(true);
                            this._keyboardController.toggleDelete(false);
                        }
                        else if (!this._longPressed && key.action === "levelSwitch") {
                            if (key.level)
                                this._setActiveLevel(key.level);
                            this._setLatched(key.level === 1 &&
                                key.iconName === "keyboard-caps-lock-symbolic");
                        }
                        this._longPressed = false;
                    });
                }
                if (key.action === "levelSwitch" &&
                    key.iconName === "keyboard-shift-symbolic") {
                    layout.shiftKeys?.push(button);
                    if (key.level === "shift") {
                        button.connect("long-press", () => {
                            this._setActiveLevel(key.level);
                            this._setLatched(true);
                            this._longPressed = true;
                        });
                    }
                }
                if (key.action === "delete") {
                    button.connect("long-press", () => this._keyboardController.toggleDelete(true));
                }
                if (key.action === "languageMenu") {
                    button.connect("long-press", () => {
                        this._popupLanguageMenu(button);
                    });
                    if (_this.kanaActive)
                        button.setLatched(_this.kanaActive);
                    _this.toggleIMKeySet?.add(button);
                }
                if (key.action === "modifier" && key.keyval) {
                    const modifierKeys = this._modifierKeys[key.keyval] || [];
                    modifierKeys.push(button);
                    this._modifierKeys[key.keyval] = modifierKeys;
                }
                if (key.action || key.keyval)
                    button.keyButton?.add_style_class_name("default-key");
                layout.appendKey(button, key.width, key.height, key.leftOffset);
            }
        };
    }
    overridePopupLanguageMenu(_originalMethod) {
        return function (keyActor) {
            if (this._languagePopup)
                this._languagePopup.destroy();
            this._languagePopup = new LanguageSelectionPopup(keyActor);
            Main.layoutManager.addTopChrome(this._languagePopup.actor);
            this._languagePopup.open(true);
        };
    }
    setupKeyboard() {
        Main.layoutManager.removeChrome(Main.layoutManager.keyboardBox);
        const destroyed = this.destroyKeyboard();
        this.getDefaultLayouts()._unregister();
        this.getModifiedLayouts()?._register();
        this.injectionManager?.overrideMethod(KeyboardBase.Keyboard.prototype, "_addRowKeys", this.overrideAddRowKeys.bind(this));
        this.injectionManager?.overrideMethod(KeyboardBase.Keyboard.prototype, "_popupLanguageMenu", this.overridePopupLanguageMenu.bind(this));
        if (destroyed) {
            Main.keyboard._keyboard = new KeyboardBase.Keyboard();
            this.allSuggestions = new AllSuggestions(this.kimpanel);
            Main.keyboard._keyboard.add_child(this.allSuggestions);
            this.allSuggestions.hide();
        }
        Main.layoutManager.addTopChrome(Main.layoutManager.keyboardBox);
        Main.keyboard._keyboard?._suggestions?.set_x_align(Clutter.ActorAlign.START);
    }
});

class KimMenu extends PopupMenu.PopupMenu {
    constructor(params) {
        const _params = Params.parse(params, {
            arrowAlignment: 0.5,
            arrowSide: St.Side.TOP,
            kimpanel: null,
            sourceActor: null,
        });
        super(_params.sourceActor, _params.arrowAlignment, _params.arrowSide);
        this.openStateChangedId = this.connect("open-state-changed", this.onOpenStateChanged.bind(this));
        this.kimKeyPressId = this.actor.connect("key-press-event", this.onSourceKeyPress.bind(this));
        this.grabbed = false;
        this.propertySwitch = [];
        this.kimpanel = _params.kimpanel;
    }
    destroy() {
        this.disconnect(this.openStateChangedId);
        this.actor.disconnect(this.kimKeyPressId);
        if (this.grabbed) {
            this.ungrab();
        }
        this.execMenu([]);
        this.kimpanel = null;
        super.destroy();
    }
    execMenu(properties) {
        for (const property of this.propertySwitch) {
            property.destroy();
        }
        this.propertySwitch = [];
        for (const property of properties) {
            const _property = parseProperty(property);
            this.addPropertyItem(_property);
        }
        if (properties.length > 0) {
            this.open(true);
        }
    }
    addPropertyItem(property) {
        const item = createMenuItem(property);
        item.menuItemActivateId = item.connect("activate", () => this.kimpanel?.triggerProperty(item.key));
        item.menuItemDestroyId = item.connect("destroy", () => {
            item.disconnect(item.menuItemActivateId);
            item.disconnect(item.menuItemDestroyId);
        });
        item.setIcon(property.icon);
        item.label.text = property.label;
        this.propertySwitch.push(item);
        this.addMenuItem(item);
    }
    grab() {
        Main.pushModal(this.actor);
        this.eventCaptureId = global.stage.connect("captured-event", this.onEventCapture.bind(this));
        this.enterEventId = global.stage.connect("enter-event", this.onHoverCapture.bind(this));
        this.leaveEventId = global.stage.connect("leave-event", this.onHoverCapture.bind(this));
        this.keyFocusNotifyId = global.stage.connect("notify::key-focus", this.onKeyFocusChanged.bind(this));
        this.grabbed = true;
    }
    onEventCapture(_actor, event) {
        if (!this.grabbed)
            return false;
        const activeMenuContains = this.actor.contains(event.get_source());
        const eventType = event.type();
        if (eventType === Clutter.EventType.BUTTON_RELEASE) {
            if (activeMenuContains) {
                return false;
            }
            this.close(false);
            return true;
        }
        if (eventType === Clutter.EventType.BUTTON_PRESS && !activeMenuContains) {
            this.close(false);
            return true;
        }
        return false;
    }
    onHoverCapture() {
        if (!this.grabbed)
            return false;
        return false;
    }
    onKeyFocusChanged() {
        if (!this.grabbed)
            return;
        const focus = global.stage.key_focus;
        if (focus) {
            if (this.actor.contains(focus))
                return;
        }
        this.close(false);
    }
    onOpenStateChanged(_menu, open) {
        if (open) {
            if (!this.grabbed)
                this.grab();
        }
        else {
            if (this.grabbed)
                this.ungrab();
        }
        const monitor = Main.layoutManager.primaryMonitor;
        if (monitor != null)
            this.actor.style = `max-height: ${Math.round(monitor.height - Main.panel.get_actor().height)}px;`;
    }
    onSourceKeyPress(_actor, event) {
        const symbol = event.get_key_symbol();
        if (symbol === Clutter.KEY_space || symbol === Clutter.KEY_Return) {
            this.toggle();
            return true;
        }
        if (symbol === Clutter.KEY_Escape && this.isOpen) {
            this.close(false);
            return true;
        }
        if (symbol === Clutter.KEY_Down) {
            if (!this.isOpen)
                this.toggle();
            this.actor.navigate_focus(this.actor, St.DirectionType.TAB_FORWARD, false);
            return true;
        }
        return false;
    }
    ungrab() {
        global.stage.disconnect(this.eventCaptureId);
        this.eventCaptureId = 0;
        global.stage.disconnect(this.enterEventId);
        this.enterEventId = 0;
        global.stage.disconnect(this.leaveEventId);
        this.leaveEventId = 0;
        global.stage.disconnect(this.keyFocusNotifyId);
        this.keyFocusNotifyId = 0;
        this.grabbed = false;
        Main.popModal(this.actor);
    }
}

class Label extends St.Label {
    buttonReleaseId;
    candidateIndex;
    enterEventId;
    ignore_focus;
    labelDestroyId;
    leaveEventId;
    touchId;
}
const createLabel = (params) => {
    const label = new St.Label(params);
    label.clutter_text.ellipsize = Pango.EllipsizeMode.NONE;
    label.clutter_text.line_wrap = false;
    return label;
};
const InputPanel = GObject.registerClass(class InputPanel extends GObject.Object {
    constructor(params) {
        super();
        const _params = Params.parse(params, { kimpanel: null });
        this.kimpanel = _params.kimpanel;
        this.arrowSide = St.Side.TOP;
        this.panel = new BoxPointer.BoxPointer(this.arrowSide, {
            x_align: Clutter.ActorAlign.START,
        });
        this.panel.style_class = "popup-menu-boxpointer";
        this.panel.add_style_class_name("popup-menu");
        this.panel.add_style_class_name("minwidth-zero");
        this.panel.add_style_class_name("kimpanel-popup-boxpointer");
        this.cursor = new St.Label({});
        this.layout = new St.BoxLayout({
            style_class: "popup-menu-content",
            vertical: true,
        });
        this.layout.add_style_class_name("kimpanel-popup-content");
        this.panel.bin.set_child(this.layout);
        this.upperLayout = new St.BoxLayout();
        this.lookupTableLayout = new St.BoxLayout({
            vertical: this.kimpanel.isLookupTableVertical(),
        });
        this.layout.add_child(this.upperLayout);
        this.text_style = this.kimpanel.getTextStyle();
        this.auxText = createLabel({
            style: this.text_style,
            style_class: "kimpanel-label",
            text: "",
        });
        this.preeditText = createLabel({
            style: this.text_style,
            style_class: "kimpanel-label",
            text: "",
        });
        this.upperLayout.add_child(this.auxText);
        this.upperLayout.add_child(this.preeditText);
        this.hide();
        this.panel.hide();
    }
    destroy() {
        if (!this.kimpanel) {
            return;
        }
        this.kimpanel = null;
        this.layout = null;
        this.upperLayout = null;
        this.lookupTableLayout = null;
        this.auxText = null;
        this.preeditText = null;
        this.panel?.destroy();
        this.panel = null;
        this.cursor?.destroy();
        this.cursor = null;
    }
    hideAux() {
        if (this.auxText?.visible)
            this.auxText.hide();
    }
    hidePreedit() {
        if (this.preeditText?.visible)
            this.preeditText.hide();
    }
    setAuxText(text) {
        this.auxText?.set_text(text);
        if (!this.auxText?.visible) {
            this.auxText?.show();
        }
    }
    setLookupTable(label, table, visible) {
        const len = visible ? table.length : 0;
        const labelLen = this.lookupTableLayout?.get_children().length ?? 0;
        if (labelLen > 0 && len === 0) {
            if (this.lookupTableLayout != null)
                this.layout?.remove_child(this.lookupTableLayout);
        }
        else if (labelLen === 0 && len > 0) {
            if (this.lookupTableLayout != null)
                this.layout?.add_child(this.lookupTableLayout);
        }
        if (len > labelLen) {
            for (let i = 0; i < len - labelLen; i++) {
                const item = createLabel({
                    reactive: true,
                    style: this.text_style,
                    style_class: "popup-menu-item kimpanel-label",
                    text: "",
                });
                item.add_style_class_name("kimpanel-candidate-item");
                item.candidateIndex = 0;
                item.ignore_focus = true;
                item.buttonReleaseId = item.connect("button-release-event", (widget) => {
                    if (!widget.ignore_focus)
                        this.candidateClicked(widget);
                });
                item.enterEventId = item.connect("enter-event", (widget) => {
                    if (!widget.ignore_focus)
                        widget.add_style_pseudo_class("hover");
                });
                item.leaveEventId = item.connect("leave-event", (widget) => {
                    if (!widget.ignore_focus)
                        widget.remove_style_pseudo_class("hover");
                });
                item.labelDestroyId = item.connect("destroy", () => {
                    if (item.buttonReleaseId != null)
                        item.disconnect(item.buttonReleaseId);
                    if (item.enterEventId != null)
                        item.disconnect(item.enterEventId);
                    if (item.leaveEventId != null)
                        item.disconnect(item.leaveEventId);
                    if (item.labelDestroyId != null)
                        item.disconnect(item.labelDestroyId);
                    if (item.touchId != null)
                        item.disconnect(item.touchId);
                });
                this.lookupTableLayout?.add_child(item);
            }
        }
        else if (len < labelLen) {
            for (let i = 0; i < labelLen - len; i++) {
                this.lookupTableLayout?.get_children()[0].destroy();
            }
        }
        const lookupTable = this.lookupTableLayout?.get_children();
        if (lookupTable == null)
            return;
        for (let i = 0; i < lookupTable.length; i++) {
            if (label[i].length === 0)
                lookupTable[i].ignore_focus = true;
            else
                lookupTable[i].ignore_focus = false;
            lookupTable[i].candidateIndex = i;
            lookupTable[i].text = label[i] + table[i];
        }
    }
    setLookupTableCursor(cursor) {
        (this.lookupTableLayout?.get_children()).forEach((label, i) => {
            if (i === cursor) {
                label.add_style_pseudo_class("active");
            }
            else {
                label.remove_style_pseudo_class("active");
            }
        });
    }
    setPreeditText(text, pos) {
        const cat = `${text.slice(0, pos)}|${text.slice(pos)}`;
        this.preeditText?.set_text(cat);
        if (!this.preeditText?.visible)
            this.preeditText?.show();
    }
    setVertical(vertical) {
        this.lookupTableLayout?.set_vertical(vertical);
    }
    updateFont(textStyle) {
        this.text_style = textStyle;
        this.auxText?.set_style(this.text_style);
        this.preeditText?.set_style(this.text_style);
        const lookupTable = this.lookupTableLayout?.get_children();
        if (lookupTable == null)
            return;
        for (const label of lookupTable) {
            label.set_style(this.text_style);
        }
    }
    updatePosition() {
        const kimpanel = this.kimpanel;
        if (kimpanel == null || this.panel == null)
            return;
        let x = kimpanel.x;
        let y = kimpanel.y;
        let w = kimpanel.w;
        let h = kimpanel.h;
        if (kimpanel.relative) {
            if (global.display.focus_window) {
                const shellScale = St.ThemeContext.get_for_stage(global.stage).scale_factor;
                const window = global.display.focus_window.get_compositor_private();
                if (window) {
                    x = window.x + x * (shellScale / kimpanel.scale);
                    y = window.y + y * (shellScale / kimpanel.scale);
                    w = w * (shellScale / kimpanel.scale);
                    h = h * (shellScale / kimpanel.scale);
                }
            }
        }
        const rect = new Mtk.Rectangle({ height: h, width: w, x: x, y: y });
        const monitor = Main.layoutManager.monitors[global.display.get_monitor_index_for_rect(rect)];
        const panel_height = this.panel?.get_height();
        if (h === 0) {
            h = 20;
            y = y - 20;
        }
        if (y + panel_height + h >= monitor.y + monitor.height) {
            this.arrowSide = St.Side.BOTTOM;
            if (y + h >= monitor.y + monitor.height) {
                y = monitor.y + monitor.height - 1;
                h = 1;
            }
        }
        else {
            this.arrowSide = St.Side.TOP;
        }
        if (x < monitor.x) {
            x = monitor.x;
        }
        if (x >= monitor.x + monitor.width) {
            x = monitor.x + monitor.width - 1;
        }
        this.cursor?.set_position(x, y);
        this.cursor?.set_size(w === 0 ? 1 : w, h === 0 ? 1 : h);
        this.panel._arrowSide = this.arrowSide;
        const visible = kimpanel.showAux || kimpanel.showPreedit || kimpanel.showLookupTable;
        if (visible) {
            this.show();
        }
        else {
            this.hide();
        }
    }
    candidateClicked(widget) {
        this.kimpanel?.selectCandidate(widget.candidate_index);
    }
    hide() {
        this.panel?.close(BoxPointer.PopupAnimation.NONE);
    }
    show() {
        if (this.cursor != null)
            this.panel?.setPosition(this.cursor, 0.0);
        this.panel?.open(BoxPointer.PopupAnimation.NONE);
        this.panel?.get_parent()?.set_child_above_sibling(this.panel, null);
    }
});

class SuggestionsManager {
    kimpanel;
    allTexts;
    cursor;
    labels;
    layoutHint;
    locked;
    texts;
    candidate = null;
    hasNext = false;
    status = "default";
    constructor(kimpanel) {
        this.kimpanel = kimpanel;
        this.allTexts = [];
        this.cursor = -1;
        this.layoutHint = 0;
        this.labels = [];
        this.locked = false;
        this.texts = [];
    }
    reset() {
        this.status = "default";
        this.candidate = null;
        this.allTexts = [];
        this.cursor = -1;
        this.layoutHint = 0;
        this.locked = false;
        this.labels = [];
        this.texts = [];
    }
    selectCandidate(candidate) {
        this.candidate = candidate;
        this.status = "search";
        this.selectCandidateHelp(this.texts);
    }
    setLookupTable(labels, texts, _attrs, hasPrev, hasNext, cursor, layout) {
        if (labels.length === 0 || labels.every((label) => label === "")) {
            this.reset();
        }
        this.hasNext = hasNext;
        this.cursor = cursor;
        this.layoutHint = layout;
        this.labels = labels;
        this.texts = texts;
        if (labels.length === 0) {
            return;
        }
        switch (this.status) {
            case "default": {
                if (this.locked)
                    return;
                if (keyboardIsVisible()) {
                    this.allTexts.push(...texts.map((text) => text.split("\n")[0]));
                    if (hasNext) {
                        this.kimpanel.lookupPageDown();
                    }
                    else {
                        this.status = "reset";
                        this.locked = true;
                        this.kimpanel.lookupPageUp();
                    }
                }
                return;
            }
            case "reset":
                if (hasPrev) {
                    this.kimpanel.lookupPageUp();
                }
                else {
                    this.status = "default";
                }
                return;
            case "search":
            case "searchReverse":
                this.selectCandidateHelp(texts);
                return;
        }
    }
    selectCandidateHelp(texts) {
        const index = texts.findIndex((value) => value.split("\n")[0] === this.candidate);
        if (index < 0) {
            if (!this.hasNext) {
                this.status = "searchReverse";
            }
            if (this.status === "searchReverse") {
                this.kimpanel.lookupPageUp();
            }
            else {
                this.kimpanel.lookupPageDown();
            }
        }
        else {
            this.kimpanel.selectCandidate(index);
            this.reset();
        }
    }
}

const FCITX_BUS_NAME = "org.fcitx.Fcitx5";
const FCITX_CONTROLLER_OBJECT_PATH = "/controller";
const FCITX_INTERFACE_CONTROLLER = "org.fcitx.Fcitx.Controller1";
const KIMPANEL_INTERFACE_INPUTMETHOD = "org.kde.kimpanel.inputmethod";
const KimpanelIface = `<node>
<interface name="org.kde.impanel">
  <signal name="MovePreeditCaret">
    <arg type="i" name="position" />
  </signal>
  <signal name="SelectCandidate">
    <arg type="i" name="index" />
  </signal>
  <signal name="LookupTablePageUp"></signal>
  <signal name="LookupTablePageDown"></signal>
  <signal name="TriggerProperty"> 
    <arg type="s" name="key" />
  </signal>
  <signal name="PanelCreated"></signal>
  <signal name="Exit"></signal>
  <signal name="ReloadConfig"></signal>
  <signal name="Configure"></signal>
  </interface>
</node>`;
const Kimpanel2Iface = `<node>
<interface name="org.kde.impanel2">
  <signal name="PanelCreated2"></signal>
  <method name="SetSpotRect">
    <arg type="i" name="x" direction="in" />
    <arg type="i" name="y" direction="in" />
    <arg type="i" name="w" direction="in" />
    <arg type="i" name="h" direction="in" />
  </method>
  <method name="SetRelativeSpotRect">
    <arg type="i" name="x" direction="in" />
    <arg type="i" name="y" direction="in" />
    <arg type="i" name="w" direction="in" />
    <arg type="i" name="h" direction="in" />
  </method>
  <method name="SetRelativeSpotRectV2">
    <arg type="i" name="x" direction="in" />
    <arg type="i" name="y" direction="in" />
    <arg type="i" name="w" direction="in" />
    <arg type="i" name="h" direction="in" />
    <arg type="d" name="scale" direction="in" />
  </method>
  <method name="SetLookupTable">
    <arg direction="in" type="as" name="label"/>
    <arg direction="in" type="as" name="text"/>
    <arg direction="in" type="as" name="attr"/>
    <arg direction="in" type="b" name="hasPrev"/>
    <arg direction="in" type="b" name="hasNext"/>
    <arg direction="in" type="i" name="cursor"/>
    <arg direction="in" type="i" name="layout"/>
  </method>
</interface>
</node>`;
const HelperIface = `<node>
<interface name="org.fcitx.GnomeHelper">
  <method name="LockXkbGroup">
    <arg direction="in" type="i" name="idx"/>
  </method>
  </interface>
</node>`;
const Kimpanel = GObject.registerClass(class Kimpanel extends GObject.Object {
    constructor(settings, dir) {
        super();
        this.isDestroyed = false;
        this.resetData();
        this.conn = Gio.bus_get_sync(Gio.BusType.SESSION, null);
        this.settings = settings;
        this.impl = Gio.DBusExportedObject.wrapJSObject(KimpanelIface, this);
        this.impl.export(Gio.DBus.session, "/org/kde/impanel");
        this.impl2 = Gio.DBusExportedObject.wrapJSObject(Kimpanel2Iface, this);
        this.impl2.export(Gio.DBus.session, "/org/kde/impanel");
        this.helperImpl = Gio.DBusExportedObject.wrapJSObject(HelperIface, this);
        this.helperImpl.export(Gio.DBus.session, "/org/fcitx/GnomeHelper");
        this.suggestionsManager = new SuggestionsManager(this);
        this.current_service = "";
        this.watch_id = 0;
        this.indicator = new KimIndicator({ kimpanel: this });
        this.inputPanel = new InputPanel({ kimpanel: this });
        this.keyboard = new Keyboard(this, dir);
        this.menu = new KimMenu({ kimpanel: this, sourceActor: this.indicator });
        this.verticalSignal = this.settings.connect("changed::vertical", () => this.inputPanel?.setVertical(this.isLookupTableVertical()));
        this.fontSignal = this.settings.connect("changed::font", () => this.inputPanel?.updateFont(this.getTextStyle()));
        this.addToShell();
        this.dbusSignal = this.conn.signal_subscribe(null, KIMPANEL_INTERFACE_INPUTMETHOD, null, null, null, Gio.DBusSignalFlags.NONE, this.parseSignal.bind(this));
        this.owner_id = Gio.bus_own_name(Gio.BusType.SESSION, "org.kde.impanel", Gio.BusNameOwnerFlags.NONE, null, () => this.requestNameFinished(), null);
        this.helper_owner_id = Gio.bus_own_name(Gio.BusType.SESSION, "org.fcitx.GnomeHelper", Gio.BusNameOwnerFlags.NONE, null, null, null);
    }
    LockXkbGroup(idx) {
        new Meta.Context().get_backend().lock_layout_group(idx);
    }
    SetLookupTable(labels, texts, attrs, hasPrev, hasNext, cursor, layout) {
        this.suggestionsManager?.setLookupTable(labels, texts, attrs, hasPrev, hasNext, cursor, layout);
        if (keyboardIsVisible()) {
            if (this.suggestionsManager != null)
                this.keyboard?.setSuggestions(this.suggestionsManager.allTexts);
        }
        else {
            this.inputPanel?.setVertical(this.isLookupTableVertical());
            this.updateInputPanel();
        }
    }
    SetRelativeSpotRect(x, y, w, h) {
        this.setRect(x, y, w, h, true, 1);
    }
    SetRelativeSpotRectV2(x, y, w, h, scale) {
        this.setRect(x, y, w, h, true, scale);
    }
    SetSpotRect(x, y, w, h) {
        this.setRect(x, y, w, h, false, 1);
    }
    destroy() {
        this.isDestroyed = true;
        this.resetData();
        this.updateInputPanel();
        if (this.watch_id !== 0) {
            Gio.bus_unwatch_name(this.watch_id);
            this.watch_id = 0;
            this.current_service = "";
        }
        this.settings?.disconnect(this.verticalSignal);
        this.settings?.disconnect(this.fontSignal);
        this.settings = null;
        this.conn?.signal_unsubscribe(this.dbusSignal);
        this.conn = null;
        Gio.bus_unown_name(this.owner_id);
        Gio.bus_unown_name(this.helper_owner_id);
        this.impl?.unexport();
        this.impl = null;
        this.impl2?.unexport();
        this.impl2 = null;
        this.helperImpl?.unexport();
        this.helperImpl = null;
        this.suggestionsManager = null;
        this.menu?.destroy();
        this.menu = null;
        this.indicator?.destroy();
        this.indicator = null;
        this.inputPanel?.destroy();
        this.inputPanel = null;
        this.keyboard?.destroy();
        this.keyboard = null;
    }
    emit(signal) {
        this.impl?.emit_signal(signal, null);
    }
    getTextStyle() {
        return getTextStyle(this.settings);
    }
    isLookupTableVertical() {
        return this.suggestionsManager?.layoutHint === 0
            ? isLookupTableVertical(this.settings)
            : this.suggestionsManager?.layoutHint === 1;
    }
    lookupPageDown() {
        this.impl?.emit_signal("LookupTablePageDown", null);
    }
    lookupPageUp() {
        this.impl?.emit_signal("LookupTablePageUp", null);
    }
    selectCandidate(arg) {
        this.impl?.emit_signal("SelectCandidate", new GLib.Variant("(i)", [arg]));
        this.suggestionsManager?.reset();
        Main.keyboard.resetSuggestions();
    }
    selectCandidateText(arg) {
        this.suggestionsManager?.selectCandidate(arg);
    }
    toggleIM() {
        this.conn?.call(FCITX_BUS_NAME, FCITX_CONTROLLER_OBJECT_PATH, FCITX_INTERFACE_CONTROLLER, "Toggle", null, null, Gio.DBusCallFlags.NONE, -1, null);
    }
    triggerProperty(arg) {
        this.impl?.emit_signal("TriggerProperty", new GLib.Variant("(s)", [arg]));
    }
    addToShell() {
        if (this.menu != null)
            Main.uiGroup.add_child(this.menu.actor);
        this.menu?.actor.hide();
        if (this.inputPanel?.panel != null)
            Main.layoutManager.addChrome(this.inputPanel.panel, {});
        if (this.inputPanel?.cursor != null)
            Main.uiGroup.add_child(this.inputPanel.cursor);
        if (this.indicator != null)
            Main.panel.addToStatusArea("kimpanel", this.indicator);
    }
    imExit(_conn, name) {
        if (this.current_service === name) {
            this.current_service = "";
            if (this.watch_id !== 0) {
                Gio.bus_unwatch_name(this.watch_id);
                this.watch_id = 0;
            }
            this.resetData();
            this.indicator?.updateProperties([]);
            this.updateInputPanel();
        }
    }
    parseSignal(_conn, sender, _object, _iface, signal, param) {
        if (this.isDestroyed) {
            return;
        }
        let changed = false;
        switch (signal) {
            case "ExecMenu": {
                const [value] = param.deepUnpack();
                this.menu?.execMenu(value);
                break;
            }
            case "RegisterProperties": {
                const [value] = param.deepUnpack();
                if (sender != null && this.current_service !== sender) {
                    this.current_service = sender;
                    if (this.watch_id !== 0) {
                        Gio.bus_unwatch_name(this.watch_id);
                    }
                    this.watch_id = Gio.bus_watch_name(Gio.BusType.SESSION, this.current_service, Gio.BusNameWatcherFlags.NONE, null, this.imExit.bind(this));
                }
                this.indicator?.updateProperties(value);
                break;
            }
            case "UpdateProperty": {
                const [value] = param.deepUnpack();
                this.indicator?.updateProperty(value);
                this.keyboard?.updateProperty(value);
                if (this.enabled)
                    this.indicator?.active();
                else
                    this.indicator?.deactive();
                break;
            }
            case "UpdateSpotLocation": {
                const value = param.deepUnpack();
                if (this.x !== value[0] ||
                    this.y !== value[1] ||
                    this.w !== 0 ||
                    this.h !== 0)
                    changed = true;
                this.x = value[0];
                this.y = value[1];
                this.w = 0;
                this.h = 0;
                break;
            }
            case "UpdatePreeditText": {
                const value = param.deepUnpack();
                if (this.preedit !== value[0])
                    changed = true;
                this.preedit = value[0];
                break;
            }
            case "UpdateAux": {
                const value = param.deepUnpack();
                if (this.aux !== value[0])
                    changed = true;
                this.aux = value[0];
                break;
            }
            case "UpdateLookupTableCursor": {
                const [value] = param.deepUnpack();
                if (this.pos !== value)
                    changed = true;
                if (this.suggestionsManager)
                    this.suggestionsManager.cursor = value;
                break;
            }
            case "UpdatePreeditCaret": {
                const [value] = param.deepUnpack();
                if (this.pos !== value)
                    changed = true;
                this.pos = value;
                break;
            }
            case "ShowPreedit": {
                const [value] = param.deepUnpack();
                if (this.showPreedit !== value)
                    changed = true;
                this.showPreedit = value;
                break;
            }
            case "ShowLookupTable": {
                const [value] = param.deepUnpack();
                if (this.showLookupTable !== value)
                    changed = true;
                this.showLookupTable = value;
                break;
            }
            case "ShowAux": {
                const [value] = param.deepUnpack();
                if (this.showAux !== value)
                    changed = true;
                this.showAux = value;
                break;
            }
            case "Enable": {
                const [value] = param.deepUnpack();
                this.enabled = value;
                if (this.enabled)
                    this.indicator?.active();
                else
                    this.indicator?.deactive();
                break;
            }
        }
        if (changed)
            this.updateInputPanel();
    }
    requestNameFinished() {
        if (this.isDestroyed) {
            return;
        }
        this.impl?.emit_signal("PanelCreated", null);
        this.impl2?.emit_signal("PanelCreated2", null);
    }
    resetData() {
        this.preedit = "";
        this.aux = "";
        this.x = 0;
        this.y = 0;
        this.w = 0;
        this.h = 0;
        this.relative = false;
        this.scale = 1;
        this.pos = 0;
        this.showPreedit = false;
        this.showLookupTable = false;
        this.showAux = false;
        this.enabled = false;
    }
    setRect(x, y, w, h, relative, scale) {
        if (this.x === x &&
            this.y === y &&
            this.w === w &&
            this.h === h &&
            this.relative === relative &&
            this.scale === scale) {
            return;
        }
        this.x = x;
        this.y = y;
        this.w = w;
        this.h = h;
        this.relative = relative;
        this.scale = scale;
        this.updateInputPanel();
    }
    updateInputPanel() {
        if (keyboardIsVisible())
            return;
        if (this.showAux) {
            this.inputPanel?.setAuxText(this.aux);
        }
        else {
            this.inputPanel?.hideAux();
        }
        if (this.showPreedit) {
            this.inputPanel?.setPreeditText(this.preedit, this.pos);
        }
        else {
            this.inputPanel?.hidePreedit();
        }
        this.inputPanel?.setLookupTable(this.suggestionsManager?.labels ?? [], this.suggestionsManager?.texts ?? [], this.showLookupTable);
        this.inputPanel?.setLookupTableCursor(this.suggestionsManager?.cursor ?? -1);
        this.inputPanel?.updatePosition();
    }
});

class IMPanelWithOSK extends Extension {
    kimpanel = null;
    settings = null;
    disable() {
        this.kimpanel?.destroy();
        this.kimpanel = null;
        this.settings = null;
    }
    enable() {
        if (this.kimpanel)
            return;
        this.settings = this.getSettings("org.gnome.shell.extensions.impanel-with-osk");
        this.kimpanel = new Kimpanel(this.settings, this.dir);
    }
}

export { IMPanelWithOSK as default };
